;;; ph-mouse.el --- Support the mouse when emacs run in a Photon pterm.

;; Copyright (C) 1998 Free Software Foundation

;; Author: Andrew Thomas <andrew@cogent.ca>
;; Keywords: mouse, terminals

;; This file is part of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Comments:

;; Enable mouse support when running inside a pterm in QNX Photon

;; It works by translating pterm escape codes into generic emacs mouse
;; events so it should work with any package that uses the mouse.

;; It uses the xterm mouse code, so don't be surprised by many xterm
;; names in this file.

;;; Todo:

;; Support multi-click -- somehow.

;; Clicking on the mode-line does not work, although it should.

;;; Code: 

(define-key global-map "\e/" nil)
(define-key esc-map "/" nil)
(define-key function-key-map "\e/" 'pterm-mouse-translate)
;;(define-key special-event-map "\e" 'pterm-mouse-event-handler)

(define-key global-map [down-mouse-3] 'mouse-drag-region)
(global-set-key [drag-mouse-3]  'mouse-set-region)

;;
;; Override mouse-movement-p to do the right thing for our returns.  I
;; have no idea how it works in xterm mode.
;;
(defsubst mouse-movement-p (object)
  "Return non-nil if OBJECT is a mouse movement event."
  ;; (insert (prin1-to-string object))
  (let ((actual (cond ((consp object) (car object))
		      ((vectorp object)
		       (car (aref object (1- (length object)))))
		      (t nil))))
    ;; (message (concat (prin1-to-string object) ": " (prin1-to-string actual)))
    (or (eq actual 'mouse-movement)
	(eq actual 'drag-mouse-1)
	(eq actual 'drag-mouse-2)
	(eq actual 'drag-mouse-3))
    )
  )

(defun pterm-read-number (termin)
  "Read a number followed by the termination character TERMIN from the
standard input."
  (let ((result 0) ch)
    (while (not (= (setq ch (read-char)) termin))
      ;;(message (concat (char-to-string ch) " not " (char-to-string termin)))
      (setq result (+ (* result 10) (- ch 48))))
    ;;(message (concat "Got " (char-to-string termin)))
    result
    )
  )

(defun pterm-mouse-translate (event)
  ;; Read a click and release event from pterm.
  (save-excursion
    (save-window-excursion
      (deactivate-mark)
      (let* ((down (pterm-mouse-event)))
	down
	)
      )
    )
  )

(defvar pterm-mouse-x 0
  "Position of last pterm mouse event relative to the frame.")

(defvar pterm-mouse-y 0
  "Position of last pterm mouse event relative to the frame.")

(defadvice mouse-position (around pterm-mouse activate)
  "Use last key from pterm-mouse-mode if available."
  (let ((answer ad-do-it))
    (setq ad-return-value
	  (if pterm-mouse-mode
	      (cons (car answer) (cons pterm-mouse-x pterm-mouse-y))
	    answer))))

(defun pterm-mouse-event-handler ()
  (interactive)
  (if (= (read-char) ?/)
      (let ((ev (pterm-mouse-event)))
	(if (not (eventp ev))
	    (message "not an event"))
	(setq unread-command-events (list ev))
	)
    )
  )

(defun pterm-mouse-event ()
  ;; Convert Pterm mouse event to Emacs mouse event.
  ;; mouse down is: \e/31;{row};{col};{type};{clicks}t
  ;; mouse up is: \e/32;{row};{col};{type};t
  ;; mouse move is: \e/33;{row};{col};{type};t
  ;; type is 1=menu, 2=adjust, 4=select
  (let* ((type (pterm-read-number 59))
	 (y (pterm-read-number 59))
	 (x (pterm-read-number 59))
	 (buttons (pterm-read-number (if (= type 31) 59 116)))
	 (clicks (if (= type 31) (pterm-read-number 116) 0))
	 (buttonno (if (= buttons 4) 1 (if (= buttons 2) 2 3)))
	 (point (cons x y))
	 (window (window-at x y))
	 (where (if window 
		    (coordinates-in-window-p point window)
		  'menu-bar))
	 (pos (if (consp where)
		  (progn
		    (select-window window)
		    (goto-char (window-start window))
		    (move-to-window-line  (cdr where))
		    (move-to-column (+ (car where) (current-column)
				       (if (string-match "\\` \\*Minibuf"
							 (buffer-name))
					   (- (minibuffer-prompt-width))
					 0)
				       (max 0 (1- (window-hscroll)))))
		    (point))
		where))
	 mevent prefevent
	 (mouse (intern (if (eq type 32)
			    (concat "mouse-" buttonno)
			  (if (eq type 31)
			      (concat "down-mouse-" buttonno)
			    (if (eq type 33)
				(concat "drag-mouse-" buttonno)))))))
    (setq pterm-mouse-x x
	  pterm-mouse-y y)
    (if (not (consp where))
	(setq prefevent (list where (list window pos point
					  (/ (nth 2 (current-time)) 1000)))))
    (setq mevent (list mouse
		       (list window pos point
			     (/ (nth 2 (current-time)) 1000))))
    (if prefevent
	(progn
	  ;; (setq unread-command-events (list prefevent mevent))
	  ;; (vector prefevent mevent)
	  (vector mevent)
	  )
      (progn
	;; (setq unread-command-events (list mevent))
	(vector mevent)
	)
      )
    ))

;; Indicator for the pterm-mouse mode.
(defvar pterm-mouse-mode nil)

(or (assq 'pterm-mouse-mode minor-mode-alist)
    (setq minor-mode-alist
	  (cons '(pterm-mouse-mode (" Mouse")) minor-mode-alist)))

;;;###autoload
(defun pterm-mouse-mode (arg)
  "Toggle Pterm mouse mode.
With prefix arg, turn Pterm mouse mode on iff arg is positive.

Turn it on to use emacs mouse commands, and off to use pterm mouse commands."
  (interactive "P")
  (if (or (and (null arg) pterm-mouse-mode)
	  (<= (prefix-numeric-value arg) 0))
      ;; Turn it off
      (if pterm-mouse-mode
	  (progn
	    (turn-off-pterm-mouse-tracking)
	    (setq pterm-mouse-mode nil)
	    (set-buffer-modified-p (buffer-modified-p))))
    ;;Turn it on
    (if pterm-mouse-mode
	()
      (setq pterm-mouse-mode t)
      (turn-on-pterm-mouse-tracking)
      (set-buffer-modified-p (buffer-modified-p)))))

(defun turn-on-pterm-mouse-tracking ()
  ;; Enable emacs mouse tracking in pterm.
  (global-set-key "\e/" nil)
  (if pterm-mouse-mode
      (send-string-to-terminal "\e/>1;6;7h\e/>9l")))

(defun turn-off-pterm-mouse-tracking ()
  ;; Disable disable emacs mouse tracking in pterm.
  (if pterm-mouse-mode
      (send-string-to-terminal "\e/0t\e/>9h")))

;; Restore normal mouse behaviour outside Emacs.
(add-hook 'suspend-hook 'turn-off-pterm-mouse-tracking)
(add-hook 'suspend-resume-hook 'turn-on-pterm-mouse-tracking)
(add-hook 'kill-emacs-hook 'turn-off-pterm-mouse-tracking)

;;; We have to reload the mouse library because we redefined the
;;; mouse-movement-p macro.  Crazily, this works.
(load-library "mouse.el")

(provide 'ph-mouse)

;;; ph-mouse.el ends here
